#!/usr/bin/env python

"""
i7index.py: Generate the index HTML file. This assumes that "index.txt",
"i7map.txt", and "template.zml" are all in the current directory.
It also uses the Python modules "agent.py" and "zsubst.py".

This script contains grumpy and recalcitrant code, which I will not attempt
to explain. "template.zml" contains both the HTML template for the output,
and all the little template-bits that are used to generate the index lines.
"""

import os
import stat
import time
import re
import agent as agentmod
import zsubst

# The URL of the manual. You could change this to the "English.lproj"
# directory of your I7 application, if you wanted the links to point to
# your local copy of the manual.
docurl = 'http://inform7.com/learn/man/'

# The directory where all the data files will be found.
basedir = '.'

# The file to be generated.
desturi = './index.html'


# To work. First, read the template file.

fl = open(os.path.join(basedir, 'template.zml'))
templatetext = fl.read()
fl.close()

template = zsubst.parsestring(templatetext)

regex_nextpart  = re.compile('("|,|==|=>|__|\*|\^|\+\+\+|`)')
regex_killbrackets = re.compile('[\\[\\]()]')

# Placeholders for entry name strings. These are used for style markup.
class Token:
    def __init__(self, label):
        self.label = label
    def __repr__(self):
        return '<%s>' % (self.label,)

CDASH = Token('CDASH')
EMPH = Token('EMPH')
FIXED = Token('FIXED')
COMMAND = Token('COMMAND')

# This is not a Token, for boring reasons.
SEEALSO = '\xA0SEEALSO'

# Represents one chapter of the manual, as read from i7map.txt.
class Chapter:
    array = []
    map = {}
    
    def __init__(self, filename, chapnum, title):
        self.filename = filename
        self.chapnum = chapnum
        if (chapnum.endswith('.')):
            chapnum = chapnum[:-1]
        self.chapsnum = chapnum
        self.title = title
        self.titlekey = title.lower()
        
        Chapter.array.append(self)
        Chapter.map[self.titlekey] = self

    def __repr__(self):
        return '<Chapter %s "%s">' % (self.chapsnum, self.title,)

# Convert an entry name list (a list of tokens and strings) into a single
# string, by dropping the tokens and pasting the strings together.
def tokenstrip(ls):
    retls = []
    for subls in ls:
        cleanls = [ val for val in subls if not isinstance(val, Token) ]
        retls.append(' '.join(cleanls))
    return retls

# Represents one index entry.
class Entry:
    array = []
    map = {}

    # Add an entry -- that is, create it (as a reference to one chapter)
    # if it doesn't already exist. If it does already exist, add the
    # additional chapter.
    @staticmethod
    def add(namels, chapname, seealso=False, isbold=False):
        name = tokenstrip(namels)
        title = ', '.join(name)
        titlekey = title.lower()
        ent = Entry.map.get(titlekey)
        if (not ent):
            ent = Entry(namels, name, title, titlekey)
            Entry.array.append(ent)
            Entry.map[titlekey] = ent
        if (not seealso):
            ent.addchapter(chapname, isbold)
        else:
            ent.setseetarget(chapname)
    
    def __init__(self, namels, name, title, titlekey):
        self.namels = namels
        self.name = name
        self.title = title
        self.titlekey = titlekey
        self.titletag = self.titlekey.replace(' ', '_').replace('"', '')

        self.isseealso = None
        self.targets = []
        self.targetstyles = []
        self.targetnames = []

    def addchapter(self, chapname, isbold=False):
        style = None
        if (isbold):
            style = EMPH
        if (self.isseealso == True):
            raise Exception('entry cannot have both "see" and a chapter: %s' % (self.title,))
        self.isseealso = False
        self.targetnames.append(chapname)
        self.targetstyles.append(style)
        self.targets.append(None)
        
    def setseetarget(self, chapname):
        if (self.isseealso == False):
            raise Exception('entry cannot have both "see" and a chapter: %s' % (self.title,))
        self.isseealso = True
        self.targetname = chapname
        
    def __repr__(self):
        return '<Entry "%s">' % (self.title,)

# Represents one group of index entries. These are hierarchical, so the
# children will be more Sections.
class Section:
    def __init__(self, label=None, namels=None):
        self.label = label
        if (label):
            self.normlabel = normalizelabel(label)
        else:
            self.normlabel = ''
        ch = self.normlabel[0:1].upper()
        if (ch < 'A' or ch > 'Z'):
            ch = '(symbol)'
        self.normletter = ch
        self.namels = namels
        self.children = {}
        self.entry = None
    def setentry(self, entry):
        if (self.entry):
            raise Exception('Section has two definitions: %s' % (self.label,))
        self.entry = entry
    def getchild(self, label):
        return self.children.get(label)
    def setchild(self, label, sec):
        self.children[label] = sec

# Normalize a string for alphabetical order.
# - Convert to lower-case, so that the sorting is case-insensitive.
# - Strip out all square brackets and parentheses, because there are entries
#   like "[if]" which should be alphabetized as "if".
# - Add a "@" to curly brackets, so that they will alphabetize before
#   letters.
def normalizelabel(str):
    str = regex_killbrackets.sub('', str)
    str = str.replace('{', '@{')
    return str.lower()

# Break up a line of input (from "index.txt") into strings and tokens,
# according to the format rules described in that file. If it encounters
# "==" or "=>", it will stop.
def parsestrings(ln):
    ls = []
    isdef = False
    start = 0

    curstr = []
    
    while (True):
        if (start >= len(ln)):
            tok = None
            break
        match = regex_nextpart.search(ln, start)
        if (not match):
            tok = None
            upto = ln[start:]
            newstart = len(ln)
        else:
            tok = match.group(1)
            upto = ln[start:match.start()]
            newstart = match.end()
        upto = upto.strip()
        if (upto):
            curstr.append(upto)
        if (tok == '"'):
            pos = ln.find('"', newstart)
            upto = ln[newstart:pos].strip()
            if (upto):
                curstr.append(upto)
            newstart = pos+1
        elif (tok == '__'):
            curstr.append(CDASH)
        elif (tok == '*'):
            curstr.append(EMPH)
        elif (tok == '`'):
            curstr.append(FIXED)
        elif (tok == '^'):
            curstr.append(COMMAND)
        elif (tok == ','):
            if (curstr):
                ls.append(curstr)
            curstr = []
        elif (tok == '+++'):
            isdef = True
        elif (tok == '=='):
            start = newstart
            break
        elif (tok == '=>'):
            if (curstr):
                ls.append(curstr)
            curstr = [SEEALSO]
            start = newstart
            tok = '=='
            break
        start = newstart
    if (curstr):
        ls.append(curstr)
    return (ls, tok, ln[start:], isdef)

# Read the "index.txt" file. Read all the index entries, and create them
# as Entry objects.
def readfile(filename):
    fl = open(filename)
    curchapname = ''
    while True:
        ln = fl.readline()
        if (not ln):
            break
        ln = ln.strip()
        if ((not ln) or (ln.startswith('#'))):
            continue
        if (ln.startswith('=')):
            ln = ln[1:].strip()
            curchapname = ln
            continue
        if (not curchapname):
            print 'Error: line appears before any chapter heading:', ln
            continue
        (ls, tok, ln2, isdef) = parsestrings(ln)
        if (tok == '=='):
            ln2 = ln2.strip()
            if (not ln2):
                print 'Error: entry name must follow ==:', ln
                continue
            Entry.add(ls, ln2, True)
            continue
        Entry.add(ls, curchapname, isbold=isdef)
    fl.close()

# Read the "i7map.txt" file. Read all the chapter entries, and create them
# as Chapter objects.
def readmap(filename):
    fl = open(filename)
    while True:
        ln = fl.readline()
        if (not ln):
            break
        ln = ln.strip()
        if ((not ln) or (ln.startswith('#'))):
            continue
        (flnm, chapnum, title) = ln.split('\t')
        chap = Chapter(flnm, chapnum, title)
    fl.close()

# Condensing the Section tree is a postprocessing step. We want to take
# any single-entry group:
#      Foo
#        bar: 1.2.3
# ...and collapse it into a single line:
#      Foo, bar: 1.2.3
# This is leaf-first recursive.
def condensetree(sec):
    for subsec in sec.children.values():
        condensetree(subsec)
        
    if (sec.entry == None and len(sec.children) == 1):
        subsec = sec.children.values()[0]
        if (subsec.label != SEEALSO):
            sec.entry = subsec.entry
            sec.children.clear()
            pos = None
            for ix in range(len(sec.namels)):
                if (not isinstance(sec.namels[ix], Token)):
                    pos = ix
            if (pos == None):
                sec.namels.append(',')
            else:
                sec.namels[pos] = sec.namels[pos] + ','
            sec.namels.extend(subsec.namels)

# The main work function. Run through all the Entries, build a Section tree,
# and return it.
def process():
    # First, look at the chapter names attached to each Entry, and convert
    # them to Chapter objects.
    for ent in Entry.array:
        if (not ent.isseealso):
            for ix in range(len(ent.targetnames)):
                name = ent.targetnames[ix]
                chap = Chapter.map.get(name.lower())
                if (not chap):
                    print 'Error: ("%s"): Chapter not found: "%s"' % (ent.title, name)
                ent.targets[ix] = chap

    # Then, build the Section tree.
    tree = Section()
    for ent in Entry.array:
        map = tree
        for (str, style) in zip(ent.name, ent.namels):
            sec = map.getchild(str)
            if (not sec):
                sec = Section(str, style)
                map.setchild(str, sec)
            map = sec
        map.setentry(ent)

    # Go through again and fix up the "see also" entries, which must now
    # refer to valid Section objects in the tree.
    for ent in Entry.array:
        if (ent.isseealso):
            val = ent.targetname
            pos = val.find(',')
            if (pos >= 0):
                val = val[ : pos ]
            ent.target = tree.getchild(val)
            if (not ent.target):
                print 'Error: ("%s"): Entry not found: "%s"' % (ent.title, ent.targetname)

    # Condense one-line groups into single lines.
    condensetree(tree)

    # Done.
    return tree

# We now need a bunch of utility functions to generate parts of the output
# file.

# Generate the list of per-letter quick links.
def letterlist(zsub, write, parentenv):
    lettertemplate = zsub.getkey(parentenv, 'indexletter')
    letterindex = zsub.getkey(parentenv, 'letterindex')
    for letter in letterindex:
        env = {}
        env['letter'] = letter
        lettertemplate.out(write, env)

# Generate the target label for a "see also" reference.
def reflist(zsub, write, parentenv):
    entrytemplate = zsub.getkey(parentenv, 'reference')
    ent = zsub.getkey(parentenv, 'entry')
    if (not ent):
        return
    isfirst = True
    for (chap, style) in zip(ent.targets, ent.targetstyles):
        if (not chap):
            continue
        env = {}
        if isfirst:
            env['isfirst'] = True
            isfirst = False
        env['chapnum'] = chap.chapnum
        env['chapsnum'] = chap.chapsnum
        env['title'] = chap.title.replace('"', '&quot;')
        env['filename'] = chap.filename
        if (style == EMPH):
            env['bold'] = True
        if (style == FIXED):
            env['fixed'] = True
        if (style == COMMAND):
            env['command'] = True
        env['docurl'] = docurl
        entrytemplate.out(write, env)

# Generate the name of an index entry. This would be easy if it weren't
# for style markup.
def namelist(zsub, write, parentenv):
    namels = zsub.getkey(parentenv, 'namels')
    style = None
    needspace = False
    for val in namels:
        if (val in [EMPH, FIXED, COMMAND]):
            if (val == style):
                if (val == EMPH):
                    write('</em>')
                elif (val == FIXED):
                    write('</code>')
                elif (val == COMMAND):
                    write('</span>')
                style = None
                needspace = True
            else:
                if (style == EMPH):
                    write('</em>')
                elif (style == FIXED):
                    write('</code>')
                elif (style == COMMAND):
                    write('</span>')
                if (needspace):
                    write(' ')
                style = val
                if (style == EMPH):
                    write('<em class="ComParam">')
                elif (style == FIXED):
                    write('<code>')
                elif (style == COMMAND):
                    write('<span class="Command">')
                needspace = False
        elif (val == CDASH):
            pass
        else:
            if (needspace):
                write(' ')
            write(val)
            needspace = True
    if (style == EMPH):
        write('</em>')
    if (style == FIXED):
        write('</code>')
    if (style == COMMAND):
        write('</span>')

# Generate the top-level list of index entries. This is also called
# recursively to generate each group and subgroup of entries.
def toplist(zsub, write, parentenv):
    entrytemplate = zsub.getkey(parentenv, 'section')
    tree = zsub.getkey(parentenv, 'tree')
    isroot = zsub.getkey(parentenv, 'isroot')
    ls = tree.children.values()
    ls.sort(key=lambda sec: sec.normlabel)
    
    if (isroot):
        lettertemplate = zsub.getkey(parentenv, 'newletter')
        curletter = None
        
    for sec in ls:
        if (isroot and sec.normletter != curletter):
            curletter = sec.normletter
            env = {}
            env['letter'] = curletter
            lettertemplate.out(write, env)
            
        label = sec.label
        env = {}
        env['label'] = label
        env['namels'] = sec.namels
        if (CDASH in sec.namels):
            env['withdash'] = True
        env['namelist'] = namelist
        if (isroot):
            env['titletag'] = label.replace(' ', '_').replace('"', '')
        if (sec.entry):
            env['entry'] = sec.entry
            if (not sec.entry.isseealso):
                env['refs'] = reflist
            else:
                if (not sec.entry.target):
                    continue
                env['see'] = True
                #env['seelabel'] = sec.entry.target.label
                env['seelabel'] = sec.entry.targetname
                env['seetag'] = sec.entry.target.label.replace(' ', '_').replace('"', '')
                if (label == SEEALSO):
                    env['seealso'] = True
                    env['namelist'] = ''
        if (sec.children):
            env['list'] = toplist
            env['tree'] = sec
        entrytemplate.out(write, env)

# A manager object for the task of generating the index output. The structure
# is defined in the "agent.py" module.
class ThisAgent(agentmod.Agent):
    def startup(self):
        readmap(os.path.join(basedir, 'i7map.txt'))
        readfile(os.path.join(basedir, 'index.txt'))
        statval = os.stat(os.path.join(basedir, 'index.txt'))
        self.uptime = time.strftime('%B %d, %Y', time.localtime(statval[stat.ST_MTIME]))
        self.tree = process()
        self.register('index')
    def geturi(self, ex, name):
        return desturi
    def generate(self, ex, name):
        ls = self.tree.children.values()
        ls.sort(key=lambda sec: sec.normlabel)
        letterindex = []
        for sec in ls:
            ch = sec.normletter
            if (letterindex and letterindex[-1] == ch):
                continue
            letterindex.append(ch)
        env = {}
        env['list'] = toplist
        env['tree'] = self.tree
        env['letterlist'] = letterlist
        env['letterindex'] = letterindex
        env['isroot'] = True
        env['uptime'] = self.uptime
        template.out(ex.write, env)

agent = ThisAgent()

if __name__ == '__main__':
    agent.startup()
    ex = agentmod.Execution()
    agent.execute(ex, 'index')
