#!/usr/bin/env python

"""
mapi7doc: Look through the I7 documentation files (in the I7 application)
and generate a table of contents. The output will be a tab-separated file,
one line per chapter. The i7index.py script reads this file (as "i7map.txt")
to generate its index output.

To run this:
    python mapi7doc.py I7_DOC_PATH > i7map.txt

The I7_DOC_PATH should be the "English.lproj" subdirectory of the Inform
application package.

This script reads the "Writing With Inform" pages (doc*.html) and the
"Recipe Book" pages (Rdoc*.html).
"""

import sys
import os
import re

if (len(sys.argv) < 2):
    print >> sys.stderr, 'Usage: python mapi7doc.py I7_DOC_PATH > i7map.txt'
    sys.exit(-1)

docdir = sys.argv[1]

regex_docfile = re.compile('doc([0-9]+)[.]html')
regex_rdocfile = re.compile('Rdoc([0-9]+)[.]html')
regex_hastitle = re.compile('<title>(.*)</title>')

MANUAL = 1
RECIPE = 2

class DocChapter:
    def __init__(self, book, index, filename, path):
        self.book = book
        self.index = (book, index)
        self.filename = filename
        self.path = path
        self.number = None
        self.title = None

chapters = []

for val in os.listdir(docdir):
    match = regex_docfile.match(val)
    if (match):
        chap = DocChapter(MANUAL, int(match.group(1)), val, os.path.join(docdir, val))
        chapters.append(chap)
    match = regex_rdocfile.match(val)
    if (match):
        chap = DocChapter(RECIPE, int(match.group(1)), val, os.path.join(docdir, val))
        chapters.append(chap)
        
if (not chapters):
    print >> sys.stderr, 'Directory does not appear to be an Inform documentation directory. (It has no "docN.html" files.) You should give the "English.lproj" subdirectory of the Inform application package.'
    sys.exit(-1)

chapters.sort(key=(lambda chap: chap.index))

for chap in chapters:
    title = None
    fl = open(chap.path)
    while True:
        ln = fl.readline()
        match = regex_hastitle.search(ln)
        if (match):
            title = match.group(1)
            break
    fl.close()
    if (not title):
        print >> sys.stderr, 'File has no title:', filename
        sys.exit(1)
    title.strip()
    title = title.replace('&lt;', '<')
    title = title.replace('&gt;', '>')
    title = title.replace('&amp;', '&')
    pos = title.find(' ')
    if (pos < 0):
        number = '?'
    else:
        number = title[:pos].strip()
        title = title[pos:].strip()
    if (chap.book == RECIPE):
        number = 'RB-' + number
        title = '(RB) ' + title
    chap.number = number
    chap.title = title

map = {}
for chap in chapters:
    title = chap.title.lower()
    ls = map.get(title)
    if (not ls):
        ls = []
        map[title] = ls
    ls.append(chap.number)

for (title, ls) in map.items():
    if (len(ls) != 1):
        print >> sys.stderr, 'Chapter title appears more than once in the documentation:', title, ls
        sys.exit(1)

sys.stdout.write('# This is a list of the chapter in the Inform 7 documentation.\n')
sys.stdout.write('# It is a simple tab-separated file.\n\n')

for chap in chapters:
    sys.stdout.write('%s\t%s\t%s\n' % (chap.filename, chap.number, chap.title))

print >> sys.stderr, 'Wrote out %d chapter titles.' % (len(chapters),)
